#!/usr/bin/ruby -Ke
##!/usr/local/ruby-1.9/bin/ruby -Ke

require 'kconv'
require 'csv'

def hash_to_sorted_array(h, level)
  if level <= 0 then
    return h
  end

  arr = h.to_a.sort {|a, b| a[0] <=> b[0]}
#  arr.reject! {|x| x[1].class == Hash} 

  for i in 0...(arr.length) do
    arr[i][1] = hash_to_sorted_array(arr[i][1], level - 1)
  end

  return arr
end


def load_csv(inf, level)
  db = Hash.new

  CSV.open(inf, 'r') do |row|
    h = db
    child = nil
    ind = -1

    if /^[0-9a-fA-F]+$/ !~ row[0] then
      $stderr.puts "Ignored line #{Kconv.kconv(row.to_s, Kconv::EUC, Kconv::SJIS)}"
      next
    end

    for i in 0...level do
      ind = row.shift.hex

      child = h[ind]
      if child == nil then
        child = Hash.new
        h[ind] = child
      end

      if i != (level - 1) then
        h = child
      end
    end

    for i in 0...(row.length) do
      if row[i] != nil then
        row[i] = Kconv.kconv(row[i], Kconv::EUC, Kconv::SJIS)
      end
    end

    h[ind] = row
  end

  return hash_to_sorted_array(db, level)
end


def calc_index_size(nindex)
  8 + 4 * nindex
end


def write_db(f, level, db, pos)
#  $stderr.puts "write_db at level #{level}"
  if level <= 0 then
    s = ""

    for i in 0...(db.length) do
#      p db
      if db[i] != nil then
        s = s + db[i].gsub(/\t/, ' ')
      end
      if i != db.length - 1 then
        s = s + "\t"
      end
    end

    ret = [0x80000000 + s.length].pack("L")
    ret = ret + s

    while (ret.length % 4) != 0 do
#      p ret.length
      ret += "\0"
    end

    f.seek(pos)
    f.write(ret)

    return ret.length

  else
    diff = db[db.length-1][0] - db[0][0]
    if diff >= 256 && db.length * 2 < diff then
      $stderr.puts "The data array too sparse; need to fix this generator"
    end

    # ΤȤ¤ʾϹθʤ
    header = [4 + 4 * (diff + 1), db[0][0]]
    offset = calc_index_size(diff + 1)
    currind = 0

    for i in db[0][0]..db[db.length-1][0] do
      if currind < db.length && i == db[currind][0] then
        header << offset
        tmp = write_db(f, level-1, db[currind][1], pos + offset)

        offset = offset + tmp
        currind = currind + 1
      else
        header << 0
      end
    end

    f.seek(pos)
    f.write(header.pack("L*"))

    return offset
  end
end


def save_db(outf, level, db)
  File.open(outf, 'w') do |f|
    write_db(f, level, db, 0)
  end
end


def main()
  inf = "in.csv"
  outf = "out.db"
  level = 1

  if ARGV.length != 3 then
    $stderr.puts "Usage: gendb.rb <level> <in.csv> <out.db>"
    exit -1
  end
  
  level = ARGV[0].to_i
  inf = ARGV[1]
  outf = ARGV[2]

  db = load_csv(inf, level)

  save_db(outf, level, db)
end

main()
