/*
 * SF-Reader: db.c
 *
 *   ̾/Ź̾ǡ١(ȴ)
 *
 *   ա ¤ʥǡ١ˤʤΤ
 *          ȡӥåǥθƤʤΤ
 *
 * (C)Copyright 2009, Yashiro Takeshi
 */

#include <basic.h>
#include <bstdlib.h>
#include <btron/file.h>
#include "db.h"

#define SFDB_RECTYPE 31
#define SFDB_RECSUBTYPE 0x7366


typedef struct DumbDatabase_ {
    W fd;
    W mapid;
    W recsize;
    B* base;
} DumbDatabase;


typedef struct DumbDBIndex_ {
    W size;
    W base;
    W indice[1]; /* dummy */
} DumbDBIndex;


typedef struct DumbDBTerminal_ {
    W size; /* ORed with 0x80000000 */
    B value[1]; /* dummy */
} DumbDBTerminal;


DumbDB* openDumbDB(LINK* lnk)
{
    W err;
    W fd;
    W recsize;
    B* addr;
    W mapid;
    DumbDatabase* db;

    /* եΥץ */
    err = opn_fil(lnk, F_READ, NULL);
    if (err < 0) {
        goto err_ret0;
    }
    fd = err;

    /* ǡ١ѥ쥳ɤ򸡺 */
    err = fnd_rec(fd, F_TOPEND, (1U<<SFDB_RECTYPE), SFDB_RECSUBTYPE, NULL);
    if (err < 0) {
        goto err_ret1;
    }

    /* 쥳ɥ */
    err = rea_rec(fd, 0, NULL, 0, &recsize, NULL);
    if (err < 0) {
        goto err_ret1;
    }

    /* 쥳ɤΥޥåԥ */
    err = map_rec(fd, 0, &addr, recsize, F_READ);
    if (err >= 0) {
        mapid = err;
    }
    else {
        mapid = -1;

        /* map_rec ˼ԤϡȤꤢɤ߹ */
        addr = (B*)malloc(recsize);
        if (addr == NULL) {
            goto err_ret1;
        }

        err = rea_rec(fd, 0, addr, recsize, NULL, NULL);
        if (err < 0) {
            goto err_ret2;
        }
    }

    /* ֥å */
    db = (DumbDatabase*)malloc(sizeof(DumbDatabase));
    if (db == NULL) {
        goto err_ret2;
    }

    memset(db, 0, sizeof(*db));
    db->fd = fd;
    db->mapid = mapid;
    db->recsize = recsize;
    db->base = addr;

    return (DumbDB*)db;

 err_ret2:
    if (mapid >= 0) {
        ump_rec(fd, mapid);
    }
    else {
        free(addr);
    }
 err_ret1:
    cls_fil(fd);
 err_ret0:
    return NULL;
}


void closeDumbDB(DumbDB* db_)
{
    DumbDatabase* db = (DumbDatabase*)db_;

    if (db == NULL) {
        return;
    }

    if (db->fd >= 0) {
        if (db->mapid >= 0) {
            ump_rec(db->fd, db->mapid);
        }
        else {
            free(db->base);
        }
        cls_fil(db->fd);
    }

    free(db);
    return;
}


DumbDBNode* getDumbDBRoot(DumbDB* db_)
{
    DumbDatabase* db = (DumbDatabase*)db_;

    if (db == NULL) {
        return NULL;
    }

    return (DumbDBNode*)(db->base);
}


Bool isDumbDBTerminal(DumbDBNode* node)
{
    DumbDBIndex* tbl = (DumbDBIndex*)node;

    if (node == NULL) {
        return True;
    }
    if ((tbl->size & 0x80000000) != 0) {
        return True;
    }

    return False;
}


DumbDBNode* searchDumbDB(DumbDBNode* node, int index)
{
    DumbDBIndex* tbl = (DumbDBIndex*)node;
    B* base = (B*)node;
    int i;

    if (node == NULL || isDumbDBTerminal(node)) {
        return NULL;
    }

    i = index - tbl->base;

    /* ǥåϰϳʤΤǡĤʤä */
    if (i < 0 || i >= tbl->size / sizeof(W) - 1) {
        return NULL;
    }

    /* եå = 0  nil ʤΤǡĤʤä */
    if (tbl->indice[i] == 0) {
        return NULL;
    }

    return (DumbDBNode*)(&base[tbl->indice[i]]);
}


int getDumbDBTerminal(DumbDBNode* node, void** element)
{
    DumbDBTerminal* term = (DumbDBTerminal*)node;

    if (node == NULL || !isDumbDBTerminal(node)) {
        return NULL;
    }

    if (element != NULL) {
        *element = (void*)(term->value);
    }
    return (term->size & 0x7fffffff);
}


#ifdef SFREAD_DBTEST
int main()
{
    W err = 0;
    DumbDB* db = NULL;
    DumbDBNode* node;
    TC dbname[] = {
        TC_FDLM, TK_S, TK_Y, TK_S, 
        TC_FDLM, TK_s, TK_u, TK_i, TK_c, TK_a, 
        TK_MINS, TK_s, TK_t, TK_a, TK_t, TK_i, TK_o, TK_n, 
        TK_PROD, TK_d, TK_b, TNULL
    };
    int indice[] = {0x0, 0xe5, 0x27}; /* ܶ */
    LINK lnk;
    int i;
    int len;
    void *elem;

    err = get_lnk(dbname, &lnk, F_NORM);
    if (err < 0) {
        printf("get_lnk %d\n", err);
        goto err_ret0;
    }

    db = openDumbDB(&lnk);
    if (db == NULL) {
        printf("openDumbDB\n");
        goto err_ret0;
    }

    node = getDumbDBRoot(db);
    if (node == NULL) {
        printf("getDumbDBRoot\n");
        goto err_ret1;
    }

    for (i = 0; i < sizeof(indice)/sizeof(indice[0]); i++) {
        printf("searching for %d at %d-th level... ", indice[i], i+1);
        node = searchDumbDB(node, indice[i]);
        if (node == NULL) {
            printf("not found\n");
            goto err_ret1;
        }
        printf("found at %08x\n", (int)node);
    }

    len = getDumbDBTerminal(node, &elem);
    if (len < 0) {
        printf("Not a terminal\n");
        goto err_ret1;
    }
    printf("Result: len=%d, data=%s\n", len, (char*)elem);

    closeDumbDB(db);
    return 0;

 err_ret1:
    closeDumbDB(db);
 err_ret0:
    return err;
}
#endif /* SFREAD_DBTEST */

