/*
 * SF-Reader: sfedy.C
 *
 *   Edyɤ߹߽
 *
 * (C)Copyright 2009, Yashiro Takeshi
 */

#include <basic.h>
#include <tcode.h>
#include <bstring.h>
#include <bassert.h>
#include <errcode.h>
#include <btron/clk.h>
#include <btron/device.h>
#include <btron/proctask.h>
#include "sfread.h"
#include "sftask.h"
#include "libpasori.h"
#include "libpasori_command.h"


#define TAB "\t"

static char edy_columns[] = {
    "" TAB
    "" TAB
    "ѳ" TAB
    "Ĺ" TAB
    ""
};

/* sum = 33*2 */
static int edy_columns_width[] = {
    15,
    15,
    10,
    10,
    16,
};

static int edy_columns_attr[] = {
    0,
    0,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_PROP,
};

static const int nedy_columns = 
    sizeof(edy_columns_width) / sizeof(edy_columns_width[0]);


static const char* sfread_edy_strtrans(uint8 trans)
{
    switch (trans) {
    case 0x20: /* ʧ */
        return "";
    case 0x02:
        return "㡼";
    case 0x04:
        return "Edyե";
    default:
        return "";
    }
}


static inline Bool sfread_is_leapyear(int y)
{
    if ((y % 400) == 0) {
        return True;
    }
    if ((y % 100) == 0) {
        return False;
    }
    return ((y % 4) == 0) ? True : False;
}


static inline int sfread_get_mdays(int y, int m)
{
    static const int mdays[12] = {
        31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31,
    };

    if (m == 2 && sfread_is_leapyear(y)) {
        return 29;
    }
    else {
        return mdays[m-1];
    }
}


/* FIXME: ˡ˸Ψ */
void sfread_add_days(DATE_TIM* dt, int days)
{
    /* FIXME: days < 0 ̤ݡ */
    assert(days >= 0);

    dt->d_day += days;

    for (;;) {
        int mdays = sfread_get_mdays(1900 + dt->d_year, dt->d_month);
        if (dt->d_day <= mdays) {
            break;
        }

        dt->d_month++;
        dt->d_day -= mdays;

        if (dt->d_month > 12) {
            dt->d_year++;
            dt->d_month = 1;
        }
    }
}


void sfread_edy_getdatetime(DATE_TIM* dt, uint32 datetime)
{
    int tmp;

    dt->d_year = 2000 - 1900;
    dt->d_month = 1;
    dt->d_day = 1;

    sfread_add_days(dt, ((datetime >> 17) & 0x7fff));

    tmp = datetime & 0x1ffff;

    dt->d_sec = tmp % 60;
    dt->d_min = (tmp / 60) % 60;
    dt->d_hour = (tmp / 3600);

    /* Ŭ˱к */
    if (dt->d_sec == 0 && dt->d_min == 0 && dt->d_hour == 24) {
        dt->d_sec = 60;
        dt->d_min = 59;
        dt->d_hour = 23;
    }
}


SFReadDoc* sfread_read_edy(felica* fd, uint16 cardtype)
{
    W err;
    int i, j;
    uint8 raw[16*PASORI_MAX_RECORDS];
    int nitems;
    SFReadDoc* doc = NULL;

    /* Edy쥳ɤɤ߹ */
    for (j = 0; j < PASORI_MAX_RECORDS; j++) {
        sftask_display_status((TC[]){L"ţɤɤ߹Ǥޤ"}, j);
        for (i = 0; i < PASORI_IO_RETRY; i++) {
            err = felica_read_without_encryption02(fd, cardtype,
                                                   0, j, &(raw[j*16]));
            if (err == 0) {
                break;
            }
            dly_tsk(PASORI_TIME_RETRY);
        }
        if (err != 0) {
            break; /* ʾǡʤäݤ */
        }
    }
    nitems = j;
    if (nitems <= 0) {
        goto err_ret0;
    }

    /* ǡǼѰ */
    doc = sfread_doc_init(edy_columns, edy_columns_width,
                          edy_columns_attr, nedy_columns);
    if (doc == NULL) {
        printf("SF-Reader: sfread_doc_init error\n");
        goto err_ret0;
    }

    /* ɤ߹쥳ɥǡ */
    for (i = nitems - 1; i >= 0; i--) {
        uint8* b;
        DATE_TIM dt;
        int32 used;
        int32 remain;
        uint32 datetime;
        char eucrow[256];
        char* row;

        b = &(raw[i*16]);

        datetime = b[4] << 24 | b[5] << 16 | b[6] << 8 | b[7];
        sfread_edy_getdatetime(&dt, datetime);

        used = b[8] << 24 | b[9] << 16 | b[10] << 8 | b[11];
        remain = b[12] << 24 | b[13] << 16 | b[14] << 8 | b[15];
			
        sprintf(eucrow, "%04d-%02d-%02d\t%02d:%02d:%02d\t%d\t%d\t%s",
                dt.d_year + 1900, dt.d_month, dt.d_day,
                dt.d_hour, dt.d_min, dt.d_sec, (int)used, (int)remain, 
                sfread_edy_strtrans(b[0]));

        row = strdup(eucrow);
        if (row == NULL) {
            printf("SF-Reader: strdup error at sfread_read_edy\n");
            goto err_ret1;
        }

        sfread_doc_appendRow(doc, row);
    }

    /* Ĺι */
    doc->sum = (int32)(raw[12] << 24 | raw[13] << 16 | raw[14] << 8 | raw[15]);

    return doc;

 err_ret1:
    sfread_doc_cleanup(doc);
 err_ret0:
    return NULL;
}


