/*
 * SF-Reader: sfnanaco.C
 *
 *   nanacoɤ߹߽
 *
 * (C)Copyright 2009, Yashiro Takeshi
 */

#include <basic.h>
#include <tcode.h>
#include <bstring.h>
#include <errcode.h>
#include <btron/device.h>
#include <btron/proctask.h>
#include "sfread.h"
#include "sftask.h"
#include "libpasori.h"
#include "libpasori_command.h"


#define TAB "\t"

static char nanaco_columns[] = {
    "" TAB
    "" TAB
    "ѳ" TAB
    "Ĺ" TAB
    ""
};

/* sum = 33*2 */
static int nanaco_columns_width[] = {
    15,
    10,
    10,
    10,
    21,
};

static int nanaco_columns_attr[] = {
    0,
    0,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_PROP,
};

static const int nnanaco_columns = 
    sizeof(nanaco_columns_width) / sizeof(nanaco_columns_width[0]);


static const char* sfread_nanaco_strtrans(uint8 trans)
{
    switch (trans) {
    case 0x35:
        return "Ѥ";
    case 0x47: /* ʧ */
        return "";
    case 0x6f:
    case 0x70:
        return "㡼";
    case 0x83:
        return "ݥȸ";
    default:
        return "";
    }
}


SFReadDoc* sfread_read_nanaco(felica* fd, uint16 cardtype)
{
    W err;
    int i, j;
    uint8 raw[16*PASORI_MAX_RECORDS];
    int nitems;
    SFReadDoc* doc = NULL;

    /* nanaco쥳ɤɤ߹ */
    for (j = 0; j < PASORI_MAX_RECORDS; j++) {
        sftask_display_status((TC[]){L"參ɤɤ߹Ǥޤ"}, j);
        for (i = 0; i < PASORI_IO_RETRY; i++) {
            err = felica_read_without_encryption02(fd, cardtype,
                                                   0, j, &(raw[j*16]));
            if (err == 0) {
                break;
            }
            dly_tsk(PASORI_TIME_RETRY);
        }
        if (err != 0) {
            break; /* ʾǡʤäݤ */
        }
    }
    nitems = j;
    if (nitems <= 0) {
        goto err_ret0;
    }

    /* ǡǼѰ */
    doc = sfread_doc_init(nanaco_columns, nanaco_columns_width,
                          nanaco_columns_attr, nnanaco_columns);
    if (doc == NULL) {
        printf("SF-Reader: sfread_doc_init error\n");
        goto err_ret0;
    }

    /* ɤ߹쥳ɥǡ */
    for (i = nitems - 1; i >= 0; i--) {
        uint8* b;
        int y,m,d,hour,min;
        uint32 datetime;
        int32 used;
        int32 remain;
        char eucrow[256];
        char* row;

        b = &(raw[i*16]);

        datetime = b[9] << 24 | b[10] << 16 | b[11] << 8 | b[12];
        y = 2000 + ((datetime >> 21) & 0x7ff);
        m = (datetime >> 17) & 0xf;
        d = (datetime >> 12) & 0x1f;
        hour = (datetime >> 6) & 0x3f;
        min = datetime & 0x3f;

        used = b[1] << 24 | b[2] << 16 | b[3] << 8 | b[4];
        remain = b[5] << 24 | b[6] << 16 | b[7] << 8 | b[8];
			
        sprintf(eucrow, "%04d-%02d-%02d\t%02d:%02d\t%d\t%d\t%s",
                y, m, d, hour, min, (int)used, (int)remain, 
                sfread_nanaco_strtrans(b[0]));

        row = strdup(eucrow);
        if (row == NULL) {
            printf("SF-Reader: strdup error at sfread_read_nanaco\n");
            goto err_ret1;
        }

        sfread_doc_appendRow(doc, row);
    }

    /* Ĺι */
    doc->sum = (int32)(raw[5] << 24 | raw[6] << 16 | raw[7] << 8 | raw[8]);

    return doc;

 err_ret1:
    sfread_doc_cleanup(doc);
 err_ret0:
    return NULL;
}


