/*
 * SF-Reader: sfsuica.C
 *
 *   Suicaɤ߹߽
 *
 * (C)Copyright 2009, Yashiro Takeshi
 */

#include <basic.h>
#include <tcode.h>
#include <bstring.h>
#include <errcode.h>
#include <btron/device.h>
#include <btron/proctask.h>
#include "sfread.h"
#include "sftask.h"
#include "db.h"
#include "libpasori.h"
#include "libpasori_command.h"


#define TAB "\t"

static char suica_columns[] = {
    "" TAB
    "" TAB
    "о" TAB
    "Ĺ" TAB
    ""
};

/* sum = 33*2 */
static int suica_columns_width[] = {
    11,
    7*2,
    7*2,
    7,
    21,
};

static int suica_columns_attr[] = {
    0,
    0,
    0,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_PROP,
};

static const int nsuica_columns = 
    sizeof(suica_columns_width) / sizeof(suica_columns_width[0]);


static DumbDB* db_station = NULL;
static DumbDB* db_shop = NULL;

void sfread_suica_cleanupDB()
{
    if (db_station != NULL) {
        closeDumbDB(db_station);
        db_station = NULL;
    }
    if (db_shop != NULL) {
        closeDumbDB(db_shop);
        db_shop = NULL;
    }
}


static DumbDB* sfread_suica_getStationDB()
{
    LINK lnk;
    W err;

    if (db_station == NULL) {
        lnk = *(info.prog);
        err = get_lnk((TC[]){L"ݣ"},
                      &lnk, F_BASED);
        if (err < 0) {
            return NULL;
        }
        
        db_station = openDumbDB(&lnk);
    }

    return db_station;
}


static DumbDB* sfread_suica_getShopDB()
{
    LINK lnk;
    W err;

    if (db_shop == NULL) {
        lnk = *(info.prog);
        err = get_lnk((TC[]){L"ݣ𡥣"},
                      &lnk, F_BASED);
        if (err < 0) {
            return NULL;
        }
        
        db_shop = openDumbDB(&lnk);
    }

    return db_shop;
}


static int sfread_suica_calcAreaCode(uint8 lineCode, uint8 regionCode)
{
    if ((lineCode & 0x80) == 0) {
        return 0;
    }
    return regionCode;
}


static const char* sfread_suica_strtrans(uint8 trans)
{
    switch (trans) {
    case 1: /* »ʧ */
    case 19: /* ʧ() */
        return ""; 
    case 2:
        return "㡼";
    case 3:
        return "";
    case 4:
    case 5:
        return "";
    case 6:
        return "";
    case 7:
        return "ȯ";
    case 8:
        return "";
    case 13:
    case 15:
        return "Х";
    case 17:
        return "ȯ";
    case 20:
        return "";
    case 21:
        return "У";
    case 31:
    case 73:
        return "";
    case 35:
        return "";
    case 70:
        return "ʪ";
    case 72:
        return "ŵ㡼";
    case 74:
        return "ʪμ";
    case 75:
        return "ʪ";
    case 198:
    case 203:
        return "ʻʪ";
    case 132:
        return "¾";
    default:
        return "";
    }
}


static int sfread_suica_findStation(DumbDBNode* root, uint8 areaCode, 
                                    uint8 lineCode, uint8 staCode, 
                                    char** val)
{
    DumbDBNode* node = root;

    if (areaCode == 0 && lineCode == 0 && staCode == 0) {
        *val = "";
        return 0;
    }

    /* ꥢɤ򸡺 */
    node = searchDumbDB(node, areaCode);
    if (node == NULL) {
        printf("SF-Read: area code %d not found\n", areaCode);
        return ER_NOEXS;
    }

    /* 襳ɤ򸡺 */
    node = searchDumbDB(node, lineCode);
    if (node == NULL) {
        printf("SF-Read: line code %02x not found\n", lineCode);
        return ER_NOEXS;
    }

    /* إɤ򸡺 */
    node = searchDumbDB(node, staCode);
    if (node == NULL) {
        printf("SF-Read: station code %02x not found\n", staCode);
        return ER_NOEXS;
    }

    return getDumbDBTerminal(node, (void**)val);
}


static int sfread_suica_findShop(DumbDBNode* root, uint8 areaCode, 
                                 uint8 termCode, uint8 lineCode, 
                                 uint8 staCode, char** val)
{
    DumbDBNode* node;
    int i;

    if (areaCode == 0 && lineCode == 0 && staCode == 0) {
        *val = "";
        return 0;
    }

    /* areaCode == 0 ξƤΥꥢ򸡺 */
    for (i = 1; i <= 4; i++) {
        if (areaCode != 0 && i != areaCode) {
            continue;
        }

        node = root;

        /* ꥢɤ򸡺 */
        node = searchDumbDB(node, i);
        if (node == NULL) {
            continue;
        }

        /* ṳ̈򸡺 */
        node = searchDumbDB(node, termCode);
        if (node == NULL) {
            continue;
        }

        /* 襳ɤ򸡺 */
        node = searchDumbDB(node, lineCode);
        if (node == NULL) {
            continue;
        }

        /* إɤ򸡺 */
        node = searchDumbDB(node, staCode);
        if (node == NULL) {
            continue;
        }

        break;
    }

    if (i > 4) {
        printf("SF-Read: target shop not found\n");
        return ER_NOEXS;
    }

    return getDumbDBTerminal(node, (void**)val);
}


SFReadDoc* sfread_read_suica(felica* fd, uint16 cardtype)
{
    W err;
    int i, j;
    uint8 raw[16*PASORI_MAX_RECORDS];
    int nitems;
    SFReadDoc* doc = NULL;

    /* Х͵ʥ쥳ɤɤ߹ */
    for (j = 0; j < PASORI_MAX_RECORDS; j++) {
        sftask_display_status((TC[]){L"ӣᥫɤɤ߹Ǥޤ"}, j);
        for (i = 0; i < PASORI_IO_RETRY; i++) {
            err = felica_read_without_encryption02(fd, cardtype,
                                                   0, j, &(raw[j*16]));
            if (err == 0) {
                break;
            }
            dly_tsk(PASORI_TIME_RETRY);
        }
        if (err != 0) {
            break; /* ʾǡʤäݤ */
        }
    }
    nitems = j;
    if (nitems <= 0) {
        goto err_ret0;
    }

    /* ǡǼѰ */
    doc = sfread_doc_init(suica_columns, suica_columns_width,
                          suica_columns_attr, nsuica_columns);
    if (doc == NULL) {
        printf("SF-Reader: sfread_doc_init error\n");
        goto err_ret0;
    }

    /* ɤ߹쥳ɥǡ */
    for (i = nitems - 1; i >= 0; i--) {
        uint8* b;
        uint16 tmp;
        int y,m,d;
        int remain;
        char eucrow[256];
        char sta_in[32] = "";
        char sta_out[32] = "";
        char notice[32] = "";
        char* row;

        b = &(raw[i*16]);
			
        tmp = b[4] * 0x100 + b[5];
        y = 2000 + ((tmp >> 9) & 0x7f);
        m = (tmp >> 5) & 0x0f;
        d = tmp & 0x1f;
        remain = b[11]*0x100+b[10];

        if (b[0] == 0xc7 || b[0] == 0xc8) {
            /* ʪΤξ */
            DumbDB* db;
            DumbDBNode* root;
            char* val;
            int vlen;
            uint8 area_in;

            int hour, min;

            /*  */
            tmp = b[6] * 0x100 + b[7];
            hour = (tmp >> 11) & 0x1f;
            min = (tmp >> 5) & 0x3f;

            /* 襳 */
            area_in = sfread_suica_calcAreaCode(b[8], b[15]);

            sprintf(notice, "%s (%02d:%02d)", 
                    sfread_suica_strtrans(b[1]), hour, min);

            /* Ź̾ */
            db = sfread_suica_getShopDB();
            if (db == NULL) {
                goto label_shop_fallback;
            }

            root = getDumbDBRoot(db);
            if (root == NULL) {
                goto label_shop_fallback;
            }

            vlen = sfread_suica_findShop(root, area_in, b[0], b[8], b[9], &val);
            if (vlen > 0) {
                sfread_tsv_getnth(sta_in, val, vlen, 0);
                strcat(sta_in, " ");
                sfread_tsv_getnth(sta_in + strlen(sta_in), val, vlen, 1);
            }

         label_shop_fallback:
            /* ǡ١ʤäν */
            if (sta_in[0] == '\0') {
                sprintf(sta_in, "%02x-%d-%02x-%02x", 
                        b[0], area_in, b[8], b[9]);
            }
        }
        else {
            /* ¾ξ; ϥХξ̤ˤ */
            DumbDB* db;
            DumbDBNode* root;
            char* val;
            int vlen;
            uint8 area_in;
            uint8 area_out;

            /*  */
            strcpy(notice, sfread_suica_strtrans(b[1]));

            /* Хξϡ̾򸡺ʤ */
            if (b[0] == 0x05) {
                goto label_station_fallback;
            }

            /* ءо */
            area_in = sfread_suica_calcAreaCode(b[6], b[15]);
            area_out = sfread_suica_calcAreaCode(b[8], b[15]);

            db = sfread_suica_getStationDB();
            if (db == NULL) {
                goto label_station_fallback;
            }

            root = getDumbDBRoot(db);
            if (root == NULL) {
                goto label_station_fallback;
            }

            vlen = sfread_suica_findStation(root, area_in, b[6], b[7], &val);
            if (vlen > 0) {
                sfread_tsv_getnth(sta_in, val, vlen, 2);
            }

            vlen = sfread_suica_findStation(root, area_in, b[8], b[9], &val);
            if (vlen > 0) {
                sfread_tsv_getnth(sta_out, val, vlen, 2);
            }

         label_station_fallback:
            /* ǡ١ʤäν */
            if (sta_in[0] == '\0'
                && (area_in != 0 || b[6] != 0 || b[7] != 0)) {
                sprintf(sta_in, "%d-%02x-%02x", 
                        area_in, b[6], b[7]);
            }
            if (sta_out[0] == '\0'
                && (area_in != 0 || b[8] != 0 || b[9] != 0)) {
                sprintf(sta_out, "%d-%02x-%02x", 
                        area_out, b[8], b[9]);
            }
        }

        sprintf(eucrow, "%04d-%02d-%02d\t%s\t%s\t%d\t%s",
                y, m, d, sta_in, sta_out, remain, notice);

        row = strdup(eucrow);
        if (row == NULL) {
            printf("SF-Reader: strdup error at sfread_read_suica\n");
            goto err_ret1;
        }

        sfread_doc_appendRow(doc, row);
    }

    /* Ĺι */
    doc->sum = raw[11]*0x100+raw[10];

    return doc;

 err_ret1:
    sfread_doc_cleanup(doc);
 err_ret0:
    return NULL;
}


