/*
 * SF-Reader: sftask.C
 *
 *   ɤ߹ѥ֥
 *
 * (C)Copyright 2009, Yashiro Takeshi
 */

#include <basic.h>
#include <tcode.h>
#include <errcode.h>
#include <bstdlib.h>
#include <btron/device.h>
#include <btron/proctask.h>
#include "sftask.h"
#include "libpasori.h"
#include "libpasori_command.h"


static W sftask_reply(W sts, SFReadDoc* data)
{
    SFReadMessage msg;

    msg.msg_type = MS_TYPE7;
    msg.msg_size = sizeof(msg) - sizeof(W)*2;
    msg.status = sts;
    msg.data = data;

    return snd_msg(0, (MESSAGE*)&msg, WAIT);
}


static W sfread_enum_syscode(felica* fd, uint16* sc)
{
    W err;
    int i;
    uint8 cmd[256];
    uint8 resp[256];
    uint8 sc_count;

    /* Request System Codeޥ */
    cmd[0] = 0x0c;
    memcpy(&cmd[1], fd->IDm, 8);

    for (i = 0; i < PASORI_IO_RETRY; i++) {
        err = pasori_write(fd->p, cmd, 9);
        if (err < 0) {
            dly_tsk(PASORI_TIME_RETRY);
            continue;
        }
        err = pasori_read(fd->p, resp, 255);
        if (err < 0) {
            dly_tsk(PASORI_TIME_RETRY);
            continue;
        }
    }
    if (err < 0) {
        return err;
    }
    if (resp[9] > 4) {
        return ER_NOEXS;
    }

    sc_count = resp[9];
    for (i = 0; i < sc_count; i++) {
        sc[i] = resp[10+(i*2)] * 0x100 + resp[11+(i*2)];
    }
    return sc_count;
}


static W sfread_get_cardtype(felica* fd)
{
    W err;
    uint8 cmd[256];
    uint8 resp[256];
    uint16 avail_serv[] = {
        0x090f, /* Хͷ */
        0x564f, /* nanaco */
        0x170f, /* Edy */
        0x680b, /* WAON */
    };
    uint16 navail_serv = sizeof(avail_serv) / sizeof(avail_serv[0]);
    int i;

    /* Request Serviceޥ */
    cmd[0] = 0x02;
    memcpy(&(cmd[1]), fd->IDm, 8);
    cmd[9] = navail_serv; /* ӥ */
    for (i = 0; i < (int)navail_serv; i++) {
        cmd[10+i*2] = L8(avail_serv[i]);
        cmd[11+i*2] = H8(avail_serv[i]);
    }

    for (i = 0; i < PASORI_IO_RETRY; i++) {
        err = pasori_write(fd->p, cmd, 10+navail_serv*2);
        if (err < 0) {
            dly_tsk(PASORI_TIME_RETRY);
            continue;
        }
        err = pasori_read(fd->p, resp, 255);
        if (err < 0) {
            dly_tsk(PASORI_TIME_RETRY);
            continue;
        }
    }
    if (err < 0) {
        return err;
    }
    if (resp[9] == 0xff) {
        return ER_NOEXS;
    }

    /* Ĵ٤ */
    for (i = 0; i < (int)resp[9]; i++) {
        uint16 t = resp[10+i*2] + resp[11+i*2] * 0x100;
        if (t != 0xffff) {
            return avail_serv[i];
        }
    }

    /* Ĥʤä */
    return ER_NOEXS;
}


static void sftask_subtask(W arg)
{
    W err;
    int i, j;
    pasori* fd_reader;
    felica* fd_card;
    uint16 sc[4];
    int sc_count;
    felica* fd_card2 = NULL;
    uint16 cardtype;
    SFReadDoc* doc = NULL;

    for (;;) {
        /* 祿鵯׵ᤵޤԤ */
        can_wup(0);
        slp_tsk(T_FOREVER);

        pdsp_msg((TC[]){L"ƣãᥫɤõƤޤ"});

        /* ɤΥץ */
        fd_reader = pasori_open(NULL);
        if (fd_reader == NULL) {
            displayError((TC[]){L"УӣңΥץ˼Ԥޤ"});
            sftask_reply(ER_NODEV, NULL);
            goto err_ret0;
        }

        /* ޥɤ */
        pasori_init(fd_reader);

        /* ɤݥ󥰤 */
        for (i = 0; i < PASORI_IO_RETRY; i++) {
            fd_card = felica_polling(fd_reader, POLLING_ANY, 0, 0);
            if (fd_card != NULL) {
                break;
            }
            dly_tsk(PASORI_TIME_RETRY);
        }
        if (fd_card == NULL) {
            displayError((TC[]){L"ͭʥɤĤޤ"});
            sftask_reply(ER_NODEV, NULL);
            goto err_ret1;
        }

        /* ƥॳɤ */
        err = sfread_enum_syscode(fd_card, sc);
        if (err < 0) {
            displayError((TC[]){L"ɤμȽ̤˼Ԥޤ"});
            sftask_reply(ER_NODEV, NULL);
            goto err_ret2;
        }
        sc_count = err;

        for (j = 0; j < sc_count; j++) {
            fd_card2 = NULL;

            /* ƥॳդǺƥݥ */
            for (i = 0; i < PASORI_IO_RETRY; i++) {
                fd_card2 = felica_polling(fd_reader, sc[j], 0, 0);
                if (fd_card2 != NULL) {
                    break;
                }
                dly_tsk(PASORI_TIME_RETRY);
            }
            if (fd_card2 == NULL) {
                err = ER_NODEV;
                continue;
            }

            /* ɤμ̤ */
            err = sfread_get_cardtype(fd_card2);
            if (err >= 0) {
                break;
            }

            free(fd_card2);
            fd_card2 = NULL;
        }
        if (err < 0) {
            displayError((TC[]){L"μΥɤˤбƤޤ"});
            sftask_reply(ER_NOEXS, NULL);
            goto err_ret3;
        }
        cardtype = err;

        switch (cardtype) {
        case 0x090f: /* Хͷ */
            doc = sfread_read_suica(fd_card2, cardtype);
            break;
        case 0x564f: /* nanaco */
            doc = sfread_read_nanaco(fd_card2, cardtype);
            break;
        case 0x170f: /* Edy */
            doc = sfread_read_edy(fd_card2, cardtype);
            break;
        case 0x680b: /* WAON */
            doc = sfread_read_waon(fd_card2, cardtype);
            break;
        default:
            displayError((TC[]){L"Υɤɤ߹ߤϼƤޤ"});
            sftask_reply(ER_NOSPT, NULL);
            goto err_ret3;
        }

        if (doc == NULL) {
            displayError((TC[]){L"ɤɤ߹ߤ˼Ԥޤ"});
            sftask_reply(ER_IO, NULL);
            goto err_ret3;
        }

        pdsp_msg(NULL);

        /* ̤ᥤ󥿥¦ */
        err = sftask_reply(ER_OK, doc);
        if (err < 0 && doc != NULL) {
            sfread_doc_cleanup(doc);
        }

        /* FALLTHROUGH */
     err_ret3:
        free(fd_card2);
        fd_card2 = NULL;
     err_ret2:
        free(fd_card);
        fd_card = NULL;
     err_ret1:
        pasori_close(fd_reader);
        fd_reader = NULL;
     err_ret0:
        continue;
    }

    info.tskid = -1;
    ext_tsk();
}


W sftask_start()
{
    W err = cre_tsk(sftask_subtask, -1, 0);
    if (err < 0) {
        return err;
    }
    info.tskid = err;
    return 0;
}


W sftask_terminate()
{
    if (info.tskid >= 0) {
        ter_tsk(info.tskid);
        info.tskid = -1;
    }

    return 0;
}


W sftask_cardread()
{
    return wup_tsk(info.tskid);
}

