/*
 * SF-Reader: sfwaon.C
 *
 *   WAONɤ߹߽
 *
 * (C)Copyright 2009, Yashiro Takeshi
 */

#include <basic.h>
#include <tcode.h>
#include <bstring.h>
#include <errcode.h>
#include <btron/device.h>
#include <btron/proctask.h>
#include "sfread.h"
#include "sftask.h"
#include "libpasori.h"
#include "libpasori_command.h"


#define TAB "\t"

static char waon_columns[] = {
    "" TAB
    "" TAB
    "ѳ" TAB
    "Ĺ" TAB
    ""
};

/* sum = 33*2 */
static int waon_columns_width[] = {
    15,
    10,
    10,
    10,
    21,
};

static int waon_columns_attr[] = {
    0,
    0,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_RIGHT,
    SFREAD_COLUMN_ATTR_PROP,
};

static const int nwaon_columns = 
    sizeof(waon_columns_width) / sizeof(waon_columns_width[0]);


static const char* sfread_waon_strtrans(uint8 trans)
{
    switch (trans) {
    case 0x04: /* ʧ */
        return "";
    case 0x0c:
    case 0x10:
        return "㡼";
    default:
        return "";
    }
}


SFReadDoc* sfread_read_waon(felica* fd, uint16 cardtype)
{
    W err;
    int i, j;
    uint8 raw[16*PASORI_MAX_RECORDS];
    int nitems;
    SFReadDoc* doc = NULL;
    uint32 sum;

    /* WAON쥳ɤɤ߹ */
    for (j = 0; j < PASORI_MAX_RECORDS; j++) {
        sftask_display_status((TC[]){L"ףϣΥɤɤ߹Ǥޤ"}, j);
        for (i = 0; i < PASORI_IO_RETRY; i++) {
            err = felica_read_without_encryption02(fd, cardtype,
                                                   0, j, &(raw[j*16]));
            if (err == 0) {
                break;
            }
            dly_tsk(PASORI_TIME_RETRY);
        }
        if (err != 0) {
            break; /* ʾǡʤäݤ */
        }
    }
    nitems = j / 2; /* WAONɤ2쥳ɤ1ȥ */
    if (nitems <= 0) {
        goto err_ret0;
    }

    /* ǡǼѰ */
    doc = sfread_doc_init(waon_columns, waon_columns_width,
                          waon_columns_attr, nwaon_columns);
    if (doc == NULL) {
        printf("SF-Reader: sfread_doc_init error\n");
        goto err_ret0;
    }

    /* ɤ߹쥳ɥǡ */
    for (i = nitems - 1; i >= 0; i--) {
        uint8* b;
        int y,m,d,hour,min;
        uint32 charged;
        uint32 datetime;
        int32 used;
        int32 remain;
        char eucrow[256];
        char* row;

        b = &(raw[i*32]);

        datetime = b[18] << 24 | b[19] << 16 | b[20] << 8 | b[21];
        datetime >>= 7;
        datetime &= 0x01ffffff;

        y = ((datetime >> 20) & 0x1f) + 2005; /* not sure */
        m = (datetime >> 16) & 0xf;
        d = (datetime >> 11) & 0x1f;
        hour = (datetime >> 6) & 0x1f;
        min = datetime & 0x3f;

        remain = b[21] << 16 | b[22] << 8 | b[23];
        remain >>= 5;
        remain &= 0x3ffff;

        used = b[23] << 16 | b[24] << 8 | b[25];
        used >>= 3;
        used &= 0x3ffff;

        charged = b[25] << 16 | b[26] << 8 | b[27];
        charged >>= 2;
        charged &= 0x1ffff;

        sprintf(eucrow, "%04d-%02d-%02d\t%02d:%02d\t%d\t%d\t%s",
                y, m, d, hour, min, (int)used - (int)charged, (int)remain, 
                sfread_waon_strtrans(b[1]));

        row = strdup(eucrow);
        if (row == NULL) {
            printf("SF-Reader: strdup error at sfread_read_waon\n");
            goto err_ret1;
        }

        sfread_doc_appendRow(doc, row);
    }

    /* Ĺι */
    sum = raw[21] << 16 | raw[22] << 8 | raw[23];
    sum >>= 5;
    sum &= 0x3ffff;
    doc->sum = sum;

    return doc;

 err_ret1:
    sfread_doc_cleanup(doc);
 err_ret0:
    return NULL;
}


